// ===============================================
// File:			glutShell.cxx
// Description:		Glut-based OpenGL-Control
// Current:			Sept 28th 1999
// Author(s):		Thomas Kern
// ===============================================

/*
 *        thk VRML-Viewer (based on OpenGL)
 *  it shows static *.wrl files based on the VRML97 standard
 *
 *          Copyright (C) 1997-2003  Thomas Kern
 *                  All Rights Reserved
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * My email adress is: 
 *   Thomas.Kern@3Dsource.de 
 * 
 * You should find these files at: my homepage
 *   http://www.3Dsource.de
 */

//=====================================================================
//	Description
//=====================================================================
// Diese Datei steuert den VRML-Viewer.
/*
	This is the GLUT based main program.

	The most function calls have english names, therefore I hope you will
	understand it without extra explaining.

	The main idea behind these file ist to explain the use of my VRML-classes
	so you can use it in your own programs.

	"static void myGLInit(int argc, char **argv)"
	scans the directory "static char  StandardDir[] = { "Modelle/" };"
	for *.wrl (VRML) files.

	"AddGround()" creates a simple ground wireframe and a coordinate system only
	for your orientation.

	A right mouse click on "Load File" calls "ReloadFromFile(int whichFile)" 
	and this function creates a VRML-object with 
	"MyObjectControl = new VRMLControl( DateiListe[whichFile] );".
	
	Please have a look at VRMLControl.*h/*.cxx for more information.

*/

//=====================================================================
//	INCLUDES
//=====================================================================
//
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "GLSettings.H"
#include "Tools.H"

#include "VRMLControl.H"
#include "Objects.H"


//=====================================================================
// Messages
//=====================================================================
//
static const char K_B = 'B';
static const char K_4 = '4';
static const char K_6 = '6';
static const char K_2 = '2';
static const char K_8 = '8';
static const char K_a = 'a';
static const char K_A = 'A';
static const char K_y = 'y';
static const char K_Y = 'Y';
static const char K_w = 'w';

#ifdef THK_GERMAN_MESSAGES
static char M_ON[]       = {"an"};
static char M_OFF[]      = {"aus"};
static char M_LIGHT[]    = {"Licht"};
static char M_LINE[]     = {"Drahtgitter"};
static char M_SMOOTH[]   = {"Gouraud Shading"};
static char M_CULL[]     = {"Backface Culling"};
static char M_SETTINGS[] = {"Einstellungen"};
static char M_LOADFILE[] = {"Modell laden"};
static char M_QUIT[]     = {"Beenden"};
static char M_STARTBENCH[] = {"Starte Benchmark-Test."};
static char M_STOPBENCH[]  = {"Stoppe Benchmark-Test."};
static char M_BENCH[]      = {"Aktuelle Framerate"};
static char M_FOG[]      = {"Nebel"};

#else
static char M_ON[]       = {"on"};
static char M_OFF[]      = {"off"};
static char M_LIGHT[]    = {"Light"};
static char M_LINE[]     = {"Wireframe"};
static char M_SMOOTH[]   = {"Gouraud Shading"};
static char M_CULL[]     = {"Backface Culling"};
static char M_SETTINGS[] = {"Settings"};
static char M_LOADFILE[] = {"Load File"};
static char M_QUIT[]     = {"Quit"};
static char M_STARTBENCH[] = {"starting benchmark"};
static char M_STOPBENCH[]  = {"benchmark stopped"};
static char M_BENCH[]      = {"current frames per second"};
static char M_FOG[]      = {"horrible Fog ;-)"};
#endif


//=====================================================================
// global Definitions
//=====================================================================
//
VRMLControl     *MyObjectControl = NULL;
GLCoordSystem	  *MyWC            = NULL;

static boolean    readyToDraw   = false;
#ifdef MESA // Voodoo
 static GLfloat    WindowX = 640.0, WindowY = 480.0;
// static GLfloat    WindowX = 820.0, WindowY = 620.0;
#else
 static GLfloat    WindowX = 400.0, WindowY = 400.0;
#endif 
//static GLdouble   WindowAspect  = 1.0;

static boolean    LightON  = true;
static boolean    FillON   = true;
static boolean    CullON   = false;
static boolean    SmoothON = true;
static boolean    WC_ON    = true;
static boolean    GroundON = false;
static boolean    FogON    = true;

static int   DateiCount    = 0;
static char  StandardDir[] = { "Modelle/" };
static char  DateiListe[MAX_DIR_ENTRIES][NAMELENGHT];

static GLfloat CarPos[] = {200.0, 200.0, 2.0};
static void Benchmark(); // forward declaration
static GLfloat xp[3], zp[3];


//=====================================================================
// Status setzen
//=====================================================================
//
// Bodengitter
static void SetGround(boolean SetON)
{
  if(SetON) {
    GroundON = true;
    //printf("  %s: \t\t %s\n", M_GROUND, M_ON);
  }
  else {
    GroundON = false;
    //printf("  %s: \t\t\t %s\n", M_GROUND, M_OFF);
  }
}

//--------------------------------------------------------------------
//
static void SwitchLighting()
{
  LightON = !LightON;
  GLSetLighting(LightON);  
}

//--------------------------------------------------------------------
//
static void SwitchDrawStile()
{
  FillON = !FillON;
  GLSetDrawStile(FillON);
}

//--------------------------------------------------------------------
//
static void SwitchCulling()
{
  CullON = !CullON;
  GLSetCulling(CullON);
}

//--------------------------------------------------------------------
//
static void SwitchShading()
{
  SmoothON = !SmoothON;
  GLSetShading(SmoothON);
}

//--------------------------------------------------------------------
//
static void SwitchGround()
{
  GroundON = !GroundON;
  SetGround(GroundON);
}

//--------------------------------------------------------------------
//
static void SwitchFog()
{
  FogON = !FogON;
  GLSetFog(FogON);
}


//=====================================================================
// Initialisierung
//=====================================================================
//
// Einstellungen auch nach Datei neuladen ermoeglichen
static void setState()
{
  //printf("%s\n", M_SETTINGS);
  GLSetLighting(LightON);
  GLSetDrawStile(FillON);
  GLSetCulling(CullON);
  GLSetShading(SmoothON);
  SetGround(GroundON);
  GLSetFog(FogON);
  GLSetEyePoint(0.0, 1.0, 80.0);
  GLSetEyeRot(0.0, 0.0, 0.0);
}

//--------------------------------------------------------------------
//
static void myGLInit(int argc, char **argv)
{
  const int   len3 = WORD_LENGHT;
  char        myDir[len3+1];
  int         len1;

  GLReset((int)WindowX, (int)WindowY);

  // Verzeichnisnamen ermitteln
  if(argc > 2 && !strcmp(argv[1],"-dir") ) {
    len1 = strlen(argv[1]);
    if( argv[2][len1-1] != '/' )
      sprintf(myDir, "%s/", argv[2]);
    else
      sprintf(myDir, "%s", argv[2]);
  }
  else
    sprintf(myDir, "%s", StandardDir);

  // Dateinamen ermitteln
  ScanDirectory(myDir, "*.wrl", DateiListe, DateiCount);
  if(!DateiCount) { // Verzeichnis enthielt keine Dateien
    strcpy(myDir, "./");
    ScanDirectory(myDir, "*.wrl", DateiListe, DateiCount);
  }

  setState();
  readyToDraw = true;
}


//=====================================================================
// VRML-Objekte laden / anlegen
//=====================================================================
//
// VRML-Objekt anlegen
static void AddGround()
{
  // MyMap     = new GLMap();
  MyWC	    = new GLCoordSystem();
}

//--------------------------------------------------------------------
// VRML-Objekt neu laden
static void ReloadFromFile(int whichFile)
{
  if(MyObjectControl) {
    delete MyObjectControl;
    MyObjectControl = NULL;
  }
  MyObjectControl = new VRMLControl( DateiListe[whichFile] );
  setState();
}


//--------------------------------------------------------------------
// VRML-Objekt neu laden
static void Quit()
{
  if(MyWC)
    delete MyWC;
  if(MyObjectControl)
    delete MyObjectControl;

  exit(EXIT_SUCCESS);
}


//=====================================================================
// control functions
//=====================================================================
//
// Keyboard events
static void myKeyboard(unsigned char key, int /*x*/, int /*y*/)
{
  if( readyToDraw ) {
    switch (key) {
    case K_B:
      Benchmark();
      break;
    case K_4:
      GLAddEyeRot(0.0, -5.0, 0.0);
      break;
    case K_6:
      GLAddEyeRot(0.0, +5.0, 0.0);
      break;
    case K_2:
      GLAddEyeRot(+5.0, 0.0, 0.0);
      break;
    case K_8:
      GLAddEyeRot(-5.0, 0.0, 0.0);
      break;
    case K_a:
      GLAddEyeMove(-0.5);
      break;
    case K_y:
      GLAddEyeMove(+0.5);
      break;
    case K_A:
      GLAddEyeMove(-2.0);
      break;
    case K_Y:
      GLAddEyeMove(+2.0);
      break;
    case K_w:
      if(WC_ON)
        WC_ON = false;
      else
	      WC_ON = true;
      break;
    case 27:
      Quit();
      break;
    case 'p':
      coordDir(xp, zp);
      printf("xp: %2.2f, %2.2f, %2.2f\n", xp[0], xp[1], xp[2]);
      printf("zp: %2.2f, %2.2f, %2.2f\n", zp[0], zp[1], zp[2]);
      break;
    }
    glutPostRedisplay();
  }
}

//--------------------------------------------------------------------
// Mouse events
static void myMouse(int x, int y)
{
  GLfloat rotX, rotY;

  if( readyToDraw && WindowX && WindowY ) {
    rotY = (GLfloat)( 180*(2*x/WindowX - 1) );
    rotX = (GLfloat)( 180*(2*y/WindowY - 1) );
    GLSetEyeRot(rotX, rotY, 0.0);
    glutPostRedisplay();
  }
}

//--------------------------------------------------------------------
// Reaktion auf Submenue
static void ReloadMenueSelect(int mode)
{
  if( mode >=0 && mode < MAX_DIR_ENTRIES )
    ReloadFromFile(mode);
  glutPostRedisplay();
}

//--------------------------------------------------------------------
// Reaktion auf Menue (Umschalten der Darstellung)
static void rightMenueSelect(int mode)
{
  switch (mode) {
  case  1:  SwitchLighting();     break;
  case  2:  SwitchDrawStile();    break;
  case  3:  SwitchCulling();      break;
  case  4:  SwitchShading();      break;
  case  6:  SwitchGround();       break;
  case  7:  SwitchFog();          break;

  case 10:  Quit();
  }
  glutPostRedisplay();
}

//--------------------------------------------------------------------
// Menue ueber Maus erreichbar
static void myGlutMenu(void)
{
  int i;
  int reloadMenu;

  // erstes SubMenu erstellen
  reloadMenu = glutCreateMenu(ReloadMenueSelect);
  for(i = 0; i < DateiCount-1; i++)
    glutAddMenuEntry( DateiListe[i], i );

  // Hauptmenue erstellen
  glutCreateMenu(rightMenueSelect);
  glutAddSubMenu(M_LOADFILE, 	reloadMenu);
  glutAddMenuEntry(M_LIGHT,     1);
  glutAddMenuEntry(M_LINE, 	    2);
  glutAddMenuEntry(M_CULL,   	  3);
  glutAddMenuEntry(M_SMOOTH,   	4);
  //glutAddMenuEntry(M_TEXTURE,   5);
  glutAddMenuEntry(M_FOG,       7);
  glutAddMenuEntry(M_QUIT,     10);

  glutAttachMenu(GLUT_RIGHT_BUTTON);
}


//=====================================================================
// Zeichnen
//=====================================================================
//
// Zeichnen
static void myDisplay(void)
{
  if(readyToDraw) {
    readyToDraw = false;

    GLnewFrame();

    if(WC_ON) {
      glDisable(GL_CULL_FACE);
      MyWC->DrawGLObjectAsList();
      if(CullON)
        glEnable(GL_CULL_FACE);
    }

    // ab hier den aktuellen Matrix-Stack auslesen
    // coordDir(xp, zp);

    //
    if(MyObjectControl) {
      if(FillON)
	      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      else
	      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      
      MyObjectControl->DrawAsList();
    }

    glutSwapBuffers();
    readyToDraw = true;
  }
}

//=====================================================================
// benchmark
//=====================================================================
//
static void Benchmark()
{
  time_t    tstart, tend;
  int       i, testvalue, count, dt, fps;
  float     temp;

  glutReshapeFunc(NULL);
  glutKeyboardFunc(NULL);
  glutPassiveMotionFunc(NULL);

  // zunaechst eine Abschaetzung der Laufzeit vornehmen
  testvalue = 150; // erstmal mit xx Durchlaeufen
  setState();      // definierter Zustand
  printf("%s\n", M_STARTBENCH);fflush(stdout);
  tstart = time(NULL);
  for(i = 0; i < testvalue; i++) {
    CarPos[0] += 0.1;
    CarPos[1] += 0.1;
    myDisplay();
  }
  tend = time(NULL);
  dt  = tend-tstart;

  if( dt < 20 ) { // Messung nur wiederholen, wenn unter 20sec
    // jetzt der scharfe Durchlauf
    temp  = 30.0/dt;       // Test soll etwa 30sec laufen
    count = (int)(testvalue*temp);
    setState();            // definierter Zustand
    i = 0;
    CarPos[0] = CarPos[1] = 10.0;

    // los geht's
    tstart = time(NULL);
    while( i < count ) {
      // Richtung 1
      while( i < 0.3*count ) {
        CarPos[0] += 0.2;
        CarPos[1] += 0.1;        
        myDisplay();
        i++;
      }
      // Richtung 2
      while( i < 0.7*count ) {  
        GLAddEyeRot(0.0, 0.4, 0.0);
        myDisplay();
        i++;
      }
      // Richtung 3
      while( i < count ) {
        CarPos[0] -= 0.1;
        CarPos[1] += 0.1;
        myDisplay();
        i++;
      }
    }
    // fertig
    tend = time(NULL);
    dt  = tend-tstart;
  }

  // Auswertung
  fps = i/dt;
  printf("%s\n", M_STOPBENCH);fflush(stdout);
  printf("For %d frames were needed %d seconds (%d fps).\n", i, dt, fps);

  glutReshapeFunc(GLSetPerspective);
  glutKeyboardFunc(myKeyboard);
  //glutPassiveMotionFunc(myMouse);
}


//=====================================================================
// main -- control
//=====================================================================
//
int main(int argc, char **argv)
{
  // Einstellmoeglichkeiten per Tastatur
  printf("\n======================================\n");
  printf("\t%s:\n", M_SETTINGS);
  printf("Direction: %c, %c, %c, %c\n", K_2, K_4, K_6, K_8);
  printf("Zoom     : %c, %c, %c, %c\n", K_a, K_A, K_y, K_Y);
  printf("Benchmark: %c\n",             K_B);
  printf("======================================\n");
  fflush(stdout);
  
  /* start of glut windowing and control functions */
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  glutInitWindowSize((int)WindowX, (int)WindowY);
  glutCreateWindow("thk: VRML-Viewer");

  myGLInit(argc, argv);
  glutDisplayFunc(myDisplay);
  glutReshapeFunc(GLSetPerspective); // Funktion aus GLsettings
  glutKeyboardFunc(myKeyboard);
  //glutPassiveMotionFunc(myMouse);
  glutMotionFunc(myMouse);
  //glutIdleFunc(myMouse);
  myGlutMenu();                      // Menue erst nach Init erzeugen !

  AddGround();

  glutMainLoop();
  /* end of glut windowing and control functions */

  return 0;             /* ANSI C requires main to return int. */
}
